# Public domain

import argparse
import binascii
import os
import urllib.request
from xml.dom import minidom

"""
Restructure:

# Read XML, download if not file doesn't exist.  Download again the first time
# a find fails.  Return info dict.
class Releases:
    def find(titleid):

Run silently unless there's a fatal error.  Only show log messages on fatal error.

Support cia files (mario DLC at offsets 3C08 (small file) and 1537AA58 (large file).

Warn if not cia, csu, 3ds, or 3dz ?

Prompt for manual input if titleid not found.

Test multiple renames at once.
"""

# Target file name format.  Files will be renamed to match this.  The extension
# is preserved.
#
# Format string:
#   firmware:
#   group:
#   serial:
#   id: release id, like 1018 for Tomodachi Life (EUR)
#   imagesize:
#   type: "3DSWare", "3DSGame", "3DSDemo", or "eShop" 
#   languages:
#   trimmedsize:
#   region: like "EUR"
#   titleid: like "0004000000008C400"
#   name: like "Tomadachi Life", converted to be a valid file name.
#   imgcrc:
#   filename: release file name, like "bbb-p-ec6p"
#   releasename:
#   publisher:
target_format = "{id} - {name} {region} {filename}"

# Where to fetch release DB.
xml_url = "http://3ds.essh.co/xml.php"

# Where to store XML file.
xml_file_name = "titles.xml"

# work from source dir
os.chdir(os.path.dirname(os.path.realpath(__file__)))

parser = argparse.ArgumentParser(description="Rename roms based on titleid.")
parser.add_argument("files", metavar="FILE", type=str, nargs="+", help="file to rename")
cmdargs = parser.parse_args()

def get_titleid(file_path):
    """
    Return titleid for the rom at file_path.

    Raises OSError if the file could not be opened.

    Raises ValueError if the file does not look like a rom (unrecognized
    titleid).

    Returns:
      string: title as hex, like 0004000000008C400
    """
    with open(file_path, "rb") as f:
        # go to titleid offset
        f.seek(0x0108)

        # Read 8-byte titleid field.
        titleid_bytes = f.read(8)

        # reverse it to match db
        titleid_bytes = titleid_bytes[::-1]

        # convert to hex
        titleid = binascii.hexlify(titleid_bytes)

        return titleid.decode("ascii").upper()

def fetch_xml():
    """
    Download the latest XML.

    Returns:
      string: the XML content as a UTF-8 string.
      None: if something goes wrong.
    """
    try:
        # fetch XML
        print("Downloading XML...")
        res = urllib.request.urlopen(xml_url)

        # read bytes
        content = res.read()
        print("Read %d bytes" % len(content))
        content = content.decode("utf-8")

        # remove shitty bytes
        content = content.replace("\x02", "")
        content = content.replace("\x0f", "")

        return content
    except:
        return None

def read_xml():
    """
    Read XML from disk.

    Returns:
      string: file content.
      None: if file does not exist.
    """
    try:
        with open(xml_file_name, "r") as f:
            content = f.read()
            return content
    except:
        return None

def write_xml(xml):
    """
    Write XML to disk.
    """
    print("Writing XML to %s" % xml_file_name)
    with open(xml_file_name, "w", errors="ignore") as f:
        f.write(xml)

def xml_is_valid(xml):
    if xml is None:
        return False
    return "<releases>" in xml

# Read XML or download if necessary.
def read_or_download_xml():
    current_xml = read_xml()
    if not xml_is_valid(current_xml):
        xml_from_internet = fetch_xml()
        if not xml_is_valid(xml_from_internet):
            if xml_from_internet is None:
                print("fatal: XML could not be downloaded")
            else:
                print("fatal: downloaded XML was invalid")

            # Fatal error condition (could not read or download)
            input("Press return to continue...")
            exit(0)

        write_xml(xml_from_internet)
        current_xml = xml_from_internet
    return current_xml

current_xml = read_or_download_xml()
xmldoc = minidom.parseString(current_xml)

def make_lookup():
    """Create a lookup table of titleid => minidom Element."""
    l = dict()
    releases = xmldoc.getElementsByTagName("release")
    for release in releases:
        release_id = "?"
        try:
            release_id = release.getElementsByTagName("id")[0].firstChild.nodeValue
        except:
            print("No release id")
            print(release.toxml())

        release_name = "?"
        try:
            release_name = release.getElementsByTagName("name")[0].firstChild.nodeValue
        except:
            print("No name")

        try:
            titleid = release.getElementsByTagName("titleid")[0].firstChild.nodeValue
            l[titleid] = release
        except:
            # some of the newest won't have titleid
            print("(Ignoring %s (%s) because no titleid...)" % (release_id, release_name))

    return l

lookup = make_lookup()

# If a titleid is not found in the XML, the latest XML is downloaded.
already_renewed = False # only renew once
def renew_xml():
    """Download XML.  Write to file if new XML is valid."""
    global already_renewed, current_xml, xmldoc, lookup

    if already_renewed:
        return
    already_renewed = True

    new_xml = fetch_xml()
    if xml_is_valid(new_xml):
        current_xml = new_xml
        write_xml(current_xml)
        xmldoc = minidom.parseString(current_xml)
        lookup = make_lookup()

def release_info(element):
    """
    Get release info from minidom <release> element.

    Args:
      element: minidom Element

    Returns:
      dict: id, name, publisher, region, langauges, group, imagesize, serial,
            titleid, imgcrc, filename, releasename, trimmedsize, firmware,
            type.
    """

    info = dict()

    keys = [
        "id", "name", "publisher", "region", "langauges", "group", "imagesize",
        "serial", "titleid", "imgcrc", "filename", "releasename",
        "trimmedsize", "firmware", "type"
    ];

    for key in keys:
        try:
            val = element.getElementsByTagName(key)[0].firstChild.nodeValue
            if key == "id":
                info["id"] = val.zfill(4)
            else:
                info[key] = val
        except Exception as e:
            #print("Error while getting %s:" % key, e)
            info[key] = ""

    return info

def sanitize_for_file_name(text):
    text = text.replace(": ", " - ")
    text = text.replace(":", " - ")
    text = text.replace("!", "")
    text = text.replace("?", "")
    return text

for file_path in cmdargs.files:
    try:
        print("Reading %s..." % file_path)
        titleid = get_titleid(file_path)
        print("titleid=%s" % titleid)

        if titleid not in lookup:
            print("titleid unknown")
            renew_xml()

            if titleid not in lookup:
                print("Ignoring file because titleid still not known")
                continue

        element = lookup[titleid]
        info = release_info(element)

        for key in info:
            info[key] = sanitize_for_file_name(info[key])

        dirname = os.path.dirname(file_path)
        basename = os.path.basename(file_path)
        fname, fext = os.path.splitext(basename)

        new_name = target_format.format(**info)

        target = os.path.join(dirname, new_name) + fext

        if "{" in target:
            print("Invalid format")
            input("Press enter to continue...")
            exit(1)

        os.rename(file_path, target)
        print("renamed to", target)
    except OSError:
        print("Ignoring file because it couldn't be opened")
    except ValueError:
        print("Ignoring file because a titleid was not found")

input("Press enter to continue...")

