# Public domain

import argparse
import binascii
import os
import re
import urllib.request
from xml.dom import minidom

# Target file name format.  Files will be renamed to match this.  The extension
# is preserved.
#
# If you're going to use name, make sure you also add something to make it
# unique, such as filename, or less ideally region or id.
#
# It doesn't work very well for Pokemon X/Y since the titleids are not unique.
#
# Format string:
#   firmware:
#   group:
#   serial:
#   id: release id, like 1018 for Tomodachi Life (EUR)
#   imagesize:
#   type: "3DSWare", "3DSGame", "3DSDemo", or "eShop" 
#   languages:
#   trimmedsize:
#   region: like "EUR"
#   titleid: like "0004000000008C400"
#   name: like "Tomadachi Life", converted to be a valid file name.
#   imgcrc:
#   filename: release file name, like "bbb-p-ec6p"
#   releasename: release file name, like "Tomodachi.Life.EUR.3DS-BigBlueBox"
#   publisher:
target_format = "{id} - {name} {region} {filename}"

# Where to fetch release DB.
xml_url = "http://3ds.essh.co/xml.php"

# Where to store XML file.
xml_file_name = "titles.xml"

# work from source dir
os.chdir(os.path.dirname(os.path.realpath(__file__)))

parser = argparse.ArgumentParser(description="Rename saves based on titleid.")
parser.add_argument("files", metavar="FILE", type=str, nargs="+", help="file to rename")
cmdargs = parser.parse_args()

def fetch_xml():
    """
    Download the latest XML.

    Returns:
      string: the XML content as a UTF-8 string.
      None: if something goes wrong.
    """
    try:
        # fetch XML
        print("Downloading XML...")
        res = urllib.request.urlopen(xml_url)

        # read bytes
        content = res.read()
        print("Read %d bytes" % len(content))
        content = content.decode("utf-8")

        # remove shitty bytes
        content = content.replace("\x02", "")
        content = content.replace("\x0f", "")

        return content
    except:
        return None

def read_xml():
    """
    Read XML from disk.

    Returns:
      string: file content.
      None: if file does not exist.
    """
    try:
        with open(xml_file_name, "r") as f:
            content = f.read()
            return content
    except:
        return None

def write_xml(xml):
    """
    Write XML to disk.
    """
    print("Writing XML to %s" % xml_file_name)
    with open(xml_file_name, "w", errors="ignore") as f:
        f.write(xml)

def xml_is_valid(xml):
    if xml is None:
        return False
    return "<releases>" in xml

# Read XML or download if necessary.
def read_or_download_xml():
    current_xml = read_xml()
    if not xml_is_valid(current_xml):
        xml_from_internet = fetch_xml()
        if not xml_is_valid(xml_from_internet):
            if xml_from_internet is None:
                print("fatal: XML could not be downloaded")
            else:
                print("fatal: downloaded XML was invalid")

            # Fatal error condition (could not read or download)
            input("Press return to continue...")
            exit(0)

        write_xml(xml_from_internet)
        current_xml = xml_from_internet
    return current_xml

def make_lookup(key, val):
    """Create lookup table of key => val"""
    l = dict()
    releases = xmldoc.getElementsByTagName("release")
    for release in releases:
        # Get release_id for debugging.
        release_id = "?"
        try:
            release_id = release.getElementsByTagName("id")[0].firstChild.nodeValue
        except:
            print("No release id")
            print(release.toxml())

        # Get release_name for debugging.
        release_name = "?"
        try:
            release_name = release.getElementsByTagName("name")[0].firstChild.nodeValue
        except:
            print("No name")

        # Extract key and value.
        try:
            key_val = release.getElementsByTagName(key)[0].firstChild.nodeValue
            if val == "release":
                val_val = release
            else:
                val_val = release.getElementsByTagName(val)[0].firstChild.nodeValue

            # special case because name is not unique (multiple ROMs can have
            # the same clean name)
            if key == "name":
                if key_val in l:
                    l[key_val].append(val_val)
                else:
                    l[key_val] = [val_val]
            else:
                l[key_val] = val_val
        except:
            print("(Ignoring %s (%s) because missing titleid or prop...)" % (release_id, release_name))

    return l

lookup = None
id_lookup = None
name_lookup = None # not unique! (multiple roms can have the same clean name)
filename_lookup = None
releasename_lookup = None

def renew_lookups():
    """Rebuild lookups (executed at start and after renewing XML)"""
    global lookup, id_lookup, name_lookup, filename_lookup, releasename_lookup
    lookup = make_lookup("titleid", "release")
    id_lookup = make_lookup("id", "titleid")
    name_lookup = make_lookup("name", "titleid")
    filename_lookup = make_lookup("filename", "titleid")
    releasename_lookup = make_lookup("releasename", "titleid")

current_xml = read_or_download_xml()
xmldoc = minidom.parseString(current_xml)
renew_lookups()

def guess_titleid(file_path):
    """
    Return titleid for the save at file_path.

    Raises OSError if the file could not be opened.

    Returns:
      string: title as hex, like "0004000000008C400"
      None: if the file name could not be guessed
    """
    basename = os.path.basename(file_path)
    fname, fext = os.path.splitext(basename)
    FNAME = fname.upper()

    # This is easy if the file is already titleid.sav
    if re.search(r"^[0-9a-fA-F]{16}$", fname):
        return fname

    # Also easy if it contains file name like "ctr-ekja".
    for filename in filename_lookup:
        if filename.upper() in FNAME:
            titleid = filename_lookup[filename]
            if titleid is not None:
                return titleid

    # Also easy if it contains release file name like "Pokemon.X.3DS-Contrast"
    for releasename in releasename_lookup:
        if releasename.upper() in FNAME:
            titleid = releasname_lookup[releasename]
            if titleid is not None:
                return titleid

    # Try to guess based on these factors
    release_id = None
    region = None
    name = None # clean name like Pokemon X
    filename = None 
    releasename = None # release file name like Pokemon.X.3DS-Contrast

    m1 = re.search(r"(EUR|USA|JPN|KOR|ITA|GER|SPA|FRA|CHT)", fname)
    if m1 is not None:
        region = m1.group(1)

    m2 = re.search(r"(\d+)", fname)
    if m2 is not None:
        release_id = int(m2.group(1))

    # It's a decent guess if the clean name matches and it either contains the
    # region or the release id.
    for name in name_lookup:
        sanitized_name = sanitize_for_file_name(name).upper()
        if sanitized_name in FNAME:
            titleids = name_lookup[name]
            print(titleids)
            for titleid in titleids:
                if "POKEMON X" in sanitized_name or "POKEMON Y" in sanitized_name:
                    return titleid

                release = lookup[titleid]
                info = release_info(release)
                if region is not None and region == info["region"]:
                    return titleid
                if release_id is not None and release_id == info["id"]:
                    return titleid

    # Now we're really in the guessing territory...

    # For most large numbers, assume it's the release ID.
    if release_id is not None:
        if release_id > 30 and release_id < 2000 and release_id != 100 and release_id != 1000:
            titleid = id_lookup[release_id]
            if titleid is not None:
                return titleid

    # @todo: look at levenscthein distance and present a menu of closest
    # matches.
    return None

# If a titleid is not found in the XML, the latest XML is downloaded.
already_renewed = False # only renew once
def renew_xml():
    """Download XML.  Write to file if new XML is valid."""
    global already_renewed, current_xml, xmldoc, lookup

    if already_renewed:
        return
    already_renewed = True

    new_xml = fetch_xml()
    if xml_is_valid(new_xml):
        current_xml = new_xml
        write_xml(current_xml)
        xmldoc = minidom.parseString(current_xml)
        renew_lookups()

def release_info(element):
    """
    Get release info from minidom <release> element.

    Args:
      element: minidom Element

    Returns:
      dict: id, name, publisher, region, langauges, group, imagesize, serial,
            titleid, imgcrc, filename, releasename, trimmedsize, firmware,
            type.
    """

    info = dict()

    keys = [
        "id", "name", "publisher", "region", "langauges", "group", "imagesize",
        "serial", "titleid", "imgcrc", "filename", "releasename",
        "trimmedsize", "firmware", "type"
    ];

    for key in keys:
        try:
            val = element.getElementsByTagName(key)[0].firstChild.nodeValue
            if key == "id":
                info["id"] = val.zfill(4)
            else:
                info[key] = val
        except Exception as e:
            #print("Error while getting %s:" % key, e)
            info[key] = ""

    return info

def sanitize_for_file_name(text):
    text = text.replace(": ", " - ")
    text = text.replace(":", " - ")
    text = text.replace("!", "")
    text = text.replace("?", "")
    return text

for file_path in cmdargs.files:
    try:
        print("Renaming %s..." % file_path)
        titleid = guess_titleid(file_path)
        print("Guessed %s" % titleid)

        if titleid is None:
            renew_xml()
            titleid = guess_titleid(file_path)

            if titleid is None:
                print("Could not figure out titleid based on name :(", file_path)
                continue

        print("titleid=%s" % titleid)

        if titleid not in lookup:
            print("titleid unknown")
            renew_xml()

            if titleid not in lookup:
                print("Ignoring file because titleid still not known")
                continue

        element = lookup[titleid]
        info = release_info(element)

        for key in info:
            info[key] = sanitize_for_file_name(info[key])

        dirname = os.path.dirname(file_path)
        basename = os.path.basename(file_path)
        fname, fext = os.path.splitext(basename)

        if fext != ".sav":
            print("Ignoring %s because it's not a .sav" % basename)
            continue

        if fname.upper() == titleid.upper():
            new_name = target_format.format(**info)
        else:
            new_name = titleid

        target = os.path.join(dirname, new_name) + fext

        if "{" in target:
            print("Invalid format")
            input("Press enter to continue...")
            exit(1)

        os.rename(file_path, target)
        print("renamed to", target)
    except OSError:
        print("Ignoring file because it couldn't be opened")
    except ValueError:
        print("Ignoring file because a titleid was not found")

input("Press enter to continue...")

